/*
								+---------------------------------+
								|                                 |
								|   *** Simple DIB wrapper ***    |
								|                                 |
								|  Copyright   -tHE SWINe- 2006  |
								|                                 |
								|         GDI_Bitmap.cpp          |
								|                                 |
								+---------------------------------+
*/

#include <windows.h>
#include "GDI_Bitmap.h"

/*
 *	CGDI_Bitmap::CGDI_Bitmap(int n_width, int n_height)
 *		- default constructor
 *		- creates DIB with a given dimensions; call b_Status() to see if constructor succeeded
 */
CGDI_Bitmap::CGDI_Bitmap(int n_width, int n_height)
	:m_n_width(n_width), m_n_height(n_height)
{

	if(!(m_p_bitmap_info = new BITMAPINFO))
		return;
	m_p_bitmap_info->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
	m_p_bitmap_info->bmiHeader.biWidth = n_width;
	m_p_bitmap_info->bmiHeader.biHeight = n_height;
	m_p_bitmap_info->bmiHeader.biPlanes = 1;
	m_p_bitmap_info->bmiHeader.biBitCount = 8 * sizeof(unsigned __int32);
	m_p_bitmap_info->bmiHeader.biCompression = BI_RGB;
	m_p_bitmap_info->bmiHeader.biSizeImage = n_width * m_n_height * sizeof(unsigned __int32);
	m_p_bitmap_info->bmiHeader.biXPelsPerMeter = 0;
	m_p_bitmap_info->bmiHeader.biYPelsPerMeter = 0;
	m_p_bitmap_info->bmiHeader.biClrUsed = 0;
	m_p_bitmap_info->bmiHeader.biClrImportant = 0;

	if(!(m_p_buffer = new unsigned __int32[n_width * n_height]))
		return;

	m_h_bitmap = CreateDIBSection(NULL, m_p_bitmap_info, DIB_RGB_COLORS, (void**)&m_p_buffer, NULL, 0);
}

/*
 *	CGDI_Bitmap::~CGDI_Bitmap()
 *		- deletes bitmap
 */
CGDI_Bitmap::~CGDI_Bitmap()
{
	DeleteObject(m_h_bitmap);
	/*if(m_p_buffer)
		delete m_p_buffer;*/ // DeleteObject apparently does that
	if(m_p_bitmap_info)
		delete m_p_bitmap_info;
}

/*
 *	void CGDI_Bitmap::CopyBuffer(unsigned __int32 *p_dest_bitmap)
 *		- copies contents of bitmap into a given buffer
 */
void CGDI_Bitmap::CopyBuffer(unsigned __int32 *p_dest_bitmap)
{
	memcpy(p_dest_bitmap, m_p_buffer, m_n_width * m_n_height * sizeof(unsigned __int32));
}

/*
 *	void CGDI_Bitmap::DrawBitmap(HDC h_dc, int x = 0, int y = 0,
 *		int n_width = -1, int n_height = -1, const unsigned __int32 *p_src_bitmap = 0)
 *		- draws bitmap into a given device, in case p_src_bitmap is not null,
 *		  it's contents are first copied into bitmap buffer
 */
void CGDI_Bitmap::DrawBitmap(HDC h_dc, int x, int y, int n_width, int n_height, const unsigned __int32 *p_src_bitmap)
{
	if(p_src_bitmap)
		memcpy(m_p_buffer, p_src_bitmap, m_n_width * m_n_height * sizeof(unsigned __int32));

	HDC h_dc_bmp;
	h_dc_bmp = CreateCompatibleDC(h_dc);
	SelectObject(h_dc_bmp, m_h_bitmap);
	if(n_width == -1 && n_height == -1)
		BitBlt(h_dc, x, y, m_n_width, m_n_height, h_dc_bmp, 0, 0, SRCCOPY);
	else
		StretchBlt(h_dc, x, y, n_width, n_height, h_dc_bmp, 0, 0, m_n_width, m_n_height, SRCCOPY);
	DeleteDC(h_dc_bmp);
}
