/*
								+---------------------------------+
								|                                 |
								|   *** Simple DIB wrapper ***    |
								|                                 |
								|  Copyright   -tHE SWINe- 2006  |
								|                                 |
								|          GDI_Bitmap.h           |
								|                                 |
								+---------------------------------+
*/

#ifndef __GDI_BITMAP_INCLUDED
#define __GDI_BITMAP_INCLUDED

/*
 *	class CGDI_Bitmap
 *		- simple class, allowing simple drawing of pixel rectangles to windows
 */
class CGDI_Bitmap {
protected:
	int m_n_width, m_n_height;
	unsigned __int32 *m_p_buffer;
	LPBITMAPINFO m_p_bitmap_info;
	HBITMAP m_h_bitmap;

public:
	/*
	 *	CGDI_Bitmap::CGDI_Bitmap(int n_width, int n_height)
	 *		- default constructor
	 *		- creates DIB with a given dimensions; call b_Status() to see if constructor succeeded
	 */
	CGDI_Bitmap(int n_width, int n_height);

	/*
	 *	CGDI_Bitmap::~CGDI_Bitmap()
	 *		- deletes bitmap
	 */
	~CGDI_Bitmap();

	/*
	 *	__inline bool CGDI_Bitmap::b_Status() const
	 *		- returns true in case constructor succeeded and bitmap object is ready, otherwise false
	 */
	__inline bool b_Status() const
	{
		return m_p_bitmap_info && m_p_buffer && m_h_bitmap != INVALID_HANDLE_VALUE;
	}

	/*
	 *	__inline int CGDI_Bitmap::n_Width() const
	 *		- returns width of bitmap
	 */
	__inline int n_Width() const
	{
		return m_n_width;
	}

	/*
	 *	__inline int CGDI_Bitmap::n_Height() const
	 *		- returns height of bitmap
	 */
	__inline int n_Height() const
	{
		return m_n_height;
	}

	/*
	 *	__inline unsigned __int32 CGDI_Bitmap::operator [](int n_index) const
	 *		- pixel-access operator
	 */
	__inline unsigned __int32 operator [](int n_index) const
	{
		return m_p_buffer[n_index];
	}

	/*
	 *	__inline unsigned __int32 &CGDI_Bitmap::operator [](int n_index)
	 *		- pixel-access operator
	 */
	__inline unsigned __int32 &operator [](int n_index)
	{
		return m_p_buffer[n_index];
	}

	/*
	 *	void CGDI_Bitmap::CopyBuffer(unsigned __int32 *p_dest_bitmap)
	 *		- copies contents of bitmap into a given buffer
	 */
	void CopyBuffer(unsigned __int32 *p_dest_bitmap);

	/*
	 *	void CGDI_Bitmap::DrawBitmap(HDC h_dc, int x = 0, int y = 0,
	 *		int n_width = -1, int n_height = -1, const unsigned __int32 *p_src_bitmap = 0)
	 *		- draws bitmap into a given device, in case p_src_bitmap is not null,
	 *		  it's contents are first copied into bitmap buffer
	 */
	void DrawBitmap(HDC h_dc, int x = 0, int y = 0,
		int n_width = -1, int n_height = -1, const unsigned __int32 *p_src_bitmap = 0);
};

#endif //__GDI_BITMAP_INCLUDED
