/*
								+---------------------------------+
								|                                 |
								|   *** simple win32 window ***   |
								|                                 |
								|  Copyright   -tHE SWINe- 2006  |
								|                                 |
								|            Main.cpp             |
								|                                 |
								+---------------------------------+
*/

#if defined(_MSC_VER) && !defined(__MWERKS__)
#pragma warning(disable:4786)
#endif
// disable warning C4786 "identifier was truncated to '255' characters in the debug information"

#include <vector>
#include "../berLame_src/CallStack.h"
#include <windows.h>
#include <commctrl.h>
#include <string>
#include <algorithm>
#include <math.h>
#include <shlobj.h>
#include "resource.h"
#include "Dialogs.h"
#include "../berLame_src/Dir.h"
#include "../berLame_src/Jpeg.h"
#include "../berLame_src/Huffman.h"
#include "../berLame_src/JpegEnc.h"
#include "GDI_Bitmap.h"
#include "Scanner.h"
#include "Material.h"

#define BIF_RETURNONLYFSDIRS 0x0001 // For finding a folder to start document searching
#define BIF_EDITBOX 0x0010 // Add an editbox to the dialog
#define BIF_VALIDATE 0x0020 // insist on valid result (or CANCEL)
#define BIF_NEWDIALOGSTYLE 0x0040 // Use the new dialog layout with the ability to resize
// Caller needs to call OleInitialize() before using this API
#define BIF_USENEWUI (BIF_NEWDIALOGSTYLE | BIF_EDITBOX)

class CMainWindow : public CLameDialog {
protected:
	std::string m_s_infolder;
	std::string m_s_outfolder;
	bool m_b_name_keep;
	std::string m_s_namepattern;
	int m_n_width, m_n_height;
	bool m_b_handle_rotated;
	bool m_b_hist_eq;
	bool m_b_conditional_hist;
	int m_n_below, m_n_above;
	bool m_b_contrast_only;
	int m_n_thumb_width, m_n_thumb_height;
	std::string m_s_template;
	int m_n_grid_w, m_n_grid_h;
	bool m_b_empty_set_style;
	std::string m_s_empty_style;
	std::string m_s_out_html;
	CGDI_Bitmap *m_p_left, *m_p_right;
	bool m_b_have_files;

	bool m_b_tolower;
	bool m_b_underline_conv;
	bool m_b_xhtml;
	bool m_b_target_blank;
	bool m_b_auto_mkdir;
	bool m_b_overwrite_confirm;
	bool m_b_breakop;
	bool m_b_to_subfolder;
	std::string m_s_subfolder; // v1.1 new features

	bool m_b_abort; // to distinguish abortion from error

public:
	CMainWindow()
		:CLameDialog(IDD_APPWINDOW), m_p_left(0), m_p_right(0)
	{
		Defaults();
	}

	void Defaults()
	{
		m_s_infolder = "E:\\Visual media\\clips\\z akc\\2005_01_10 - fastwell 2.3\\";
		m_s_outfolder = "E:\\Frag\\gallmak\\";
		m_b_name_keep = false;
		m_s_namepattern = "dvojka_trojka\\img_%04d.jpg";
		m_n_width = 1024;
		m_n_height = 768;
		m_b_handle_rotated = true;
		m_b_hist_eq = false;
		m_b_conditional_hist = true;
		m_n_below = 220;
		m_n_above = 30;
		m_b_contrast_only = true;
		m_n_thumb_width = 200;
		m_n_thumb_height = 150;
		m_s_template =
			"<!doctype html public \"-//w3c//dtd html 4.01 transitional//en\">\r\n"
			"<html>\r\n"
			"<head>\r\n"
			"<meta name=\"generator\" content=\"gallerymaker\">\r\n"
			"<meta http-equiv=\"content-type\" content=\"text/html; charset=windows-1250\">\r\n"
			"<title>my gallery, page $page-no of $page-num</title>\r\n"
			"</head>\r\n"
			"<body>\r\n"
			"<h2>my gallery, page $page-no of $page-num</h2>\r\n"
			"\r\n"
			"<table>\r\n"
			"$gallery-content\r\n"
			"</table>\r\n"
			"\r\n"
			"<ul>\r\n"
			"<li><a href=\"$first-file\">first page</a></li>\r\n"
			"<li><a href=\"$prev-file\">prev page</a></li>\r\n"
			"<li><a href=\"$next-file\">next page</a></li>\r\n"
			"<li><a href=\"$last-file\">last page</a></li>\r\n"
			"</ul>\r\n"
			"</body>\r\n"
			"</html>\r\n";
		m_n_grid_w = 5;
		m_n_grid_h = 15;
		m_b_empty_set_style = false;
		m_s_empty_style = "gal_empty_cell";
		m_s_out_html = "gallery_FastWell_%04d.htm";
		m_b_have_files = false;

		m_b_tolower = false;
		m_b_underline_conv = true;
		m_b_xhtml = false;
		m_b_target_blank = false;
		m_b_auto_mkdir = false;
		m_b_overwrite_confirm = true;
		m_b_to_subfolder = true;
		m_s_subfolder = "img"; // v1.1 new features
	}
	
	~CMainWindow()
	{
		if(m_p_left)
			delete m_p_left;
		if(m_p_right)
			delete m_p_right;
	}

	void SetText(int n_resource, const std::string &r_s_str) const
	{
		SetDlgItemText(m_h_dlg, n_resource, r_s_str.c_str());
	}

	void SetCheck(int n_resource, bool b_check) const
	{
		SendMessage(GetDlgItem(m_h_dlg, n_resource),
			BM_SETCHECK, (WPARAM)((b_check)? BST_CHECKED : BST_UNCHECKED), 0);
	}

	void SetInt(int n_resource, int n_value) const
	{
		SetDlgItemInt(m_h_dlg, n_resource, n_value, true);
	}

	bool GetText(int n_resource, std::string &r_s_str)
	{
		r_s_str.erase(r_s_str.begin(), r_s_str.end());
		int n_length = GetWindowTextLength(GetDlgItem(m_h_dlg, n_resource));
		r_s_str.reserve(n_length);
		if(r_s_str.capacity() < n_length)
			return false;
		r_s_str.insert(r_s_str.begin(), n_length, ' '); // don't forget!
		GetDlgItemText(m_h_dlg, n_resource, &*r_s_str.begin(), r_s_str.capacity());
		return true;
	}

	bool GetInt(int n_resource, int &r_n_value)
	{
		BOOL b_success;
		r_n_value = (int)GetDlgItemInt(m_h_dlg, n_resource, &b_success, true);
		return b_success != 0;
	}

	bool GetCheck(int n_resource)
	{
		return SendMessage(GetDlgItem(m_h_dlg, n_resource), BM_GETCHECK, 0, 0) == BST_CHECKED;
	}

	void GetCheck(int n_resource, bool &r_b_value)
	{
		r_b_value = SendMessage(GetDlgItem(m_h_dlg, n_resource), BM_GETCHECK, 0, 0) == BST_CHECKED;
	}

	void Update_Dialog() const
	{
		SetText(IDC_INFOLDER, m_s_infolder);
		SetText(IDC_OUTFOLDER, m_s_outfolder);
		SetCheck(IDC_NAME_KEEP, m_b_name_keep);
		SetCheck(IDC_NAME_RENAME, !m_b_name_keep);
		SetText(IDC_NAMEPATTERN, m_s_namepattern);
		SetInt(IDC_WIDTH, m_n_width);
		SetInt(IDC_HEIGHT, m_n_height);
		SetCheck(IDC_HANDLE_ROTATED, m_b_handle_rotated);
		SetCheck(IDC_HISTO_EQ, m_b_hist_eq);
		SetCheck(IDC_CONDITIONAL_HIST, m_b_conditional_hist);
		SetInt(IDC_BELOW, m_n_below);
		SetInt(IDC_ABOVE, m_n_above);
		SetCheck(IDC_CONTRAST_ONLY, m_b_contrast_only);
		SetCheck(IDC_COLORS, !m_b_contrast_only);
		SetInt(IDC_THUMB_WIDTH, m_n_thumb_width);
		SetInt(IDC_THUMB_HEIGHT, m_n_thumb_height);
		SetText(IDC_TEMPLATE, m_s_template);
		SetInt(IDC_GRID_W, m_n_grid_w);
		SetInt(IDC_GRID_H, m_n_grid_h);
		SetCheck(IDC_EMPTY_SET_STYLE, m_b_empty_set_style);
		SetText(IDC_EMPTY_STYLE, m_s_empty_style);
		SetText(IDC_OUT_HTML, m_s_out_html);
		SetCheck(IDC_SKIP_IMAGES, m_b_have_files);

		SetCheck(IDC_TOLOWER, m_b_tolower);
		SetCheck(IDC_UNDERLINE_CONV, m_b_underline_conv);
		SetCheck(IDC_XHTML, m_b_xhtml);
		SetCheck(IDC_TARGET_BLANK, m_b_target_blank);
		SetCheck(IDC_AUTO_MKDIR, m_b_auto_mkdir);
		SetCheck(IDC_OVERWRITE_CONFIRM, m_b_overwrite_confirm);
		SetCheck(IDC_SUBFOLDER, m_b_to_subfolder);
		SetText(IDC_SUBFOLDER_NAME, m_s_subfolder);
		// v1.1 new features
	}

	bool Update_Data()
	{
		if(!GetText(IDC_INFOLDER, m_s_infolder)) return false;
		if(!GetText(IDC_OUTFOLDER, m_s_outfolder)) return false;
		GetCheck(IDC_NAME_KEEP, m_b_name_keep);
		if(!GetText(IDC_NAMEPATTERN, m_s_namepattern)) return false;
		GetInt(IDC_WIDTH, m_n_width);
		GetInt(IDC_HEIGHT, m_n_height);
		GetCheck(IDC_HANDLE_ROTATED, m_b_handle_rotated);
		GetCheck(IDC_HISTO_EQ, m_b_hist_eq);
		GetCheck(IDC_CONDITIONAL_HIST, m_b_conditional_hist);
		GetInt(IDC_BELOW, m_n_below);
		GetInt(IDC_ABOVE, m_n_above);
		GetCheck(IDC_CONTRAST_ONLY, m_b_contrast_only);
		GetInt(IDC_THUMB_WIDTH, m_n_thumb_width);
		GetInt(IDC_THUMB_HEIGHT, m_n_thumb_height);
		if(!GetText(IDC_TEMPLATE, m_s_template)) return false;
		GetInt(IDC_GRID_W, m_n_grid_w);
		GetInt(IDC_GRID_H, m_n_grid_h);
		GetCheck(IDC_EMPTY_SET_STYLE, m_b_empty_set_style);
		if(!GetText(IDC_EMPTY_STYLE, m_s_empty_style)) return false;
		if(!GetText(IDC_OUT_HTML, m_s_out_html)) return false;
		GetCheck(IDC_SKIP_IMAGES, m_b_have_files);

		GetCheck(IDC_TOLOWER, m_b_tolower);
		GetCheck(IDC_UNDERLINE_CONV, m_b_underline_conv);
		GetCheck(IDC_XHTML, m_b_xhtml);
		GetCheck(IDC_TARGET_BLANK, m_b_target_blank);
		GetCheck(IDC_AUTO_MKDIR, m_b_auto_mkdir);
		GetCheck(IDC_OVERWRITE_CONFIRM, m_b_overwrite_confirm);
		GetCheck(IDC_SUBFOLDER, m_b_to_subfolder);
		if(!GetText(IDC_SUBFOLDER_NAME, m_s_subfolder)) return false;
		// v1.1 new features

		return true;
	}

protected:
	static int CALLBACK BrowseCallbackProc(HWND h_wnd, UINT n_msg, LPARAM n_l_param, LPARAM p_data)
	{
		switch(n_msg) {
			case BFFM_INITIALIZED: 
			if(p_data) {
				const char *p_s_folder = (const char*)p_data;
				SendMessage(h_wnd, BFFM_SETSELECTION, TRUE, (LPARAM)p_s_folder);
			}
			break;
		}
		return 0;
	}

	static bool BrowseForDirectory(HWND h_wnd_owner, std::string &r_directory, const char *p_s_name)
	{
		BROWSEINFO t_browse_info = {0};
		t_browse_info.hwndOwner = h_wnd_owner;
		t_browse_info.ulFlags = BIF_NEWDIALOGSTYLE | BIF_RETURNONLYFSDIRS;
		t_browse_info.lpfn = BrowseCallbackProc;
		t_browse_info.lParam = (LPARAM)r_directory.c_str();
		t_browse_info.lpszTitle = p_s_name;
		LPITEMIDLIST p_item_id_list = SHBrowseForFolder(&t_browse_info);
		bool b_result = false;
		if(p_item_id_list) {
			char p_s_path[MAX_PATH];
			if(SHGetPathFromIDList(p_item_id_list, p_s_path)) {
				r_directory = p_s_path;
				b_result = true;
			}
			// get the name of the folder

			IMalloc *p_imalloc = 0;
			if(SUCCEEDED(SHGetMalloc(&p_imalloc))) {
				p_imalloc->Free(p_item_id_list);
				p_imalloc->Release();
			}
			// free memory used
		}

		return b_result;
	}

	virtual int DlgProc(HWND h_dlg, unsigned int n_msg,
		unsigned int n_w_param, long n_l_param)
	{
		switch(n_msg) {
		case WM_INITDIALOG:
			if(!LoadConfig())
				Defaults();
			Update_Dialog();
			// update window data

			if(CoInitialize(NULL) != S_OK)
				MessageBox(m_h_dlg, "Error: failed to initalize \'browse for folder\' dialogs", "GalleryMaker", MB_OK);

			{
				RECT t_rect;
				GetClientRect(GetDlgItem(m_h_dlg, IDC_BEFORE), &t_rect);
				if(m_p_left = new CGDI_Bitmap(t_rect.right - t_rect.left, t_rect.bottom - t_rect.top)) {
					for(int i = 0, n = m_p_left->n_Width() * m_p_left->n_Height(); i < n; ++ i)
						(*m_p_left)[i] = n_Grey(rand() / 4);
				}
				GetClientRect(GetDlgItem(m_h_dlg, IDC_AFTER), &t_rect);
				if(m_p_right = new CGDI_Bitmap(t_rect.right - t_rect.left, t_rect.bottom - t_rect.top)) {
					for(int i = 0, n = m_p_right->n_Width() * m_p_right->n_Height(); i < n; ++ i)
						(*m_p_right)[i] = n_Grey(rand() / 4);
				}
			}

			{
				RECT t_rect;
				GetClientRect(m_h_dlg, &t_rect);
				InvalidateRect(m_h_dlg, &t_rect, false);
			}
			// repaint us
			return true;

		case WM_PAINT:
			DrawThumbnail(true, 0);
			DrawThumbnail(false, 0);
			return 0;

		case WM_COMMAND:
			if(n_w_param == IDOK) {
				EnableWindows(false); // lock the dialog
				m_b_breakop = false;
				m_b_abort = false;
				if(!Proceed()) {
					EnableWindows(true);
					if(m_b_abort)
						return 0;
					if(m_b_breakop)
						MessageBox(h_dlg, "Operation was interrupted by user", "GalleryMaker", MB_OK);
					else
						MessageBox(h_dlg, "There were errors during gallery generation", "GalleryMaker", MB_OK);
					return 0;
				}
				EnableWindows(true);
			} else if(n_w_param == IDC_DEFAULTS) {
				if(MessageBox(h_dlg, "This will reset all settings, do you really want that?", "GalleryMaker", MB_YESNO) == IDYES) {
					Defaults();
					Update_Dialog();
				}
			} else if(n_w_param == IDC_BREAK)
				m_b_breakop = true;
			else if(n_w_param == IDC_BROWSE_SRC) {
				if(BrowseForDirectory(m_h_dlg, m_s_infolder, "\nChoose directory containing input images\n"
				   /*"blah blah ... why is there so much empty space?\n"
				   "a bee flies ... it flies forth ... straight over the tennis court"*/))
					Update_Dialog();
			} if(n_w_param == IDC_BROWSE_DST) {
				if(BrowseForDirectory(m_h_dlg, m_s_outfolder, "Choose directory for output images"))
					Update_Dialog();
			} else if(n_w_param == IDCANCEL) {
				if(!SaveConfig())
					MessageBox(h_dlg, "Failed to save settings", "GalleryMaker", MB_OK);
				End();
			}
			return 0;
		}
		return 0;
	}

	static std::string s_EscapeStr(const std::string &r_s_str)
	{
		const char *p_escape_pairs[] = {
			"\\", "\\\\",
			"\'", "\\\'",
			"\"", "\\\"",
			"\t", "\\t",
			"\n", "\\n",
			"\r", "\\r"
		};
		std::string s_out;
		for(int i = 0, n = r_s_str.length(); i < n; ++ i) {
			int c = r_s_str[i];
			bool b_escaped = false;
			for(int j = 0; j < sizeof(p_escape_pairs) / sizeof(p_escape_pairs[0]) / 2; ++ j) {
				if(c == p_escape_pairs[2 * j][0]) {
					s_out += p_escape_pairs[2 * j + 1];
					b_escaped = true;
					break;
				}
			}
			if(!b_escaped)
				s_out += c;
		}

		return s_out;
	}

	bool SaveConfig()
	{
		if(!Update_Data())
			return false;

		FILE *p_fw;
		if(!(p_fw = fopen("config.ini", "w")))
			return false;

		fprintf(p_fw, "input-dir: \"%s\";\n", s_EscapeStr(m_s_infolder).c_str());
		fprintf(p_fw, "output-dir: \"%s\";\n", s_EscapeStr(m_s_outfolder).c_str());
		fprintf(p_fw, "keep-name: %s;\n", (m_b_name_keep)? "true" : "false");
		fprintf(p_fw, "name-pattern: \"%s\";\n", s_EscapeStr(m_s_namepattern).c_str());
		fprintf(p_fw, "max-width: %d;\n", m_n_width);
		fprintf(p_fw, "max-height: %d;\n", m_n_height);
		fprintf(p_fw, "handle-rotated: %s;\n", (m_b_handle_rotated)? "true" : "false");
		fprintf(p_fw, "histogram-enable: %s;\n", (m_b_hist_eq)? "true" : "false");
		fprintf(p_fw, "histogram-conditional: %s;\n", (m_b_conditional_hist)? "true" : "false");
		fprintf(p_fw, "max-feetspace: %d;\n", m_n_below);
		fprintf(p_fw, "max-headspace: %d;\n", m_n_above);
		fprintf(p_fw, "enhance-contrast-only: %s;\n", (m_b_contrast_only)? "true" : "false");
		fprintf(p_fw, "max-thumb-width: %d;\n", m_n_thumb_width);
		fprintf(p_fw, "max-thumb-height: %d;\n", m_n_thumb_height);
		fprintf(p_fw, "html-template: <![CDATA[%s]]>;\n", m_s_template.c_str());
		fprintf(p_fw, "grid-width: %d;\n", m_n_grid_w);
		fprintf(p_fw, "grid-height: %d;\n", m_n_grid_h);
		fprintf(p_fw, "style-empty-cell: %s;\n", (m_b_empty_set_style)? "true" : "false");
		fprintf(p_fw, "empty-style-name: \'%s\';\n", m_s_empty_style.c_str());
		fprintf(p_fw, "output-file: \"%s\";\n", s_EscapeStr(m_s_out_html).c_str());
		fprintf(p_fw, "skip-images: %s;\n", (m_b_have_files)? "true" : "false");

		fprintf(p_fw, "to-lowcase: %s;\n", (m_b_tolower)? "true" : "false");
		fprintf(p_fw, "underline-conversion: %s;\n", (m_b_underline_conv)? "true" : "false");
		fprintf(p_fw, "use-xhtml: %s;\n", (m_b_xhtml)? "true" : "false");
		fprintf(p_fw, "target-blank: %s;\n", (m_b_target_blank)? "true" : "false");
		fprintf(p_fw, "auto-mkdir: %s;\n", (m_b_auto_mkdir)? "true" : "false");
		fprintf(p_fw, "confirm-overwrite: %s;\n", (m_b_overwrite_confirm)? "true" : "false");
		fprintf(p_fw, "to-subfolder: %s;\n", (m_b_to_subfolder)? "true" : "false");
		fprintf(p_fw, "subfolder-name: \"%s\";\n", s_EscapeStr(m_s_subfolder).c_str());
		// v1.1 new features

		fclose(p_fw);

		return true;
	}

	bool _Parse(CMaterialLexer &scanner)
	{
		for(;;) {
			if(!scanner.Get_NextToken())
				return false;
			if(scanner.n_Cur_Token() == CMaterialLexer::token_EOF)
				break;
			else if(scanner.n_Cur_Token() != CMaterialLexer::token_Node)
				return false;

			std::string node_name = scanner.t_Cur_Token().p_s_Value();
			if(!scanner.Get_NextToken(CMaterialLexer::token_Colon) ||
			   !scanner.Get_NextToken())
				return false;

			if(!strcmp(node_name.c_str(), "input-dir")) {
				if(scanner.n_Cur_Token() != CMaterialLexer::token_String)
					return false;
				m_s_infolder = scanner.t_Cur_Token().p_s_Value();
			} else if(!strcmp(node_name.c_str(), "output-dir")) {
				if(scanner.n_Cur_Token() != CMaterialLexer::token_String)
					return false;
				m_s_outfolder = scanner.t_Cur_Token().p_s_Value();
			} else if(!strcmp(node_name.c_str(), "keep-name")) {
				if(scanner.n_Cur_Token() != CMaterialLexer::token_Bool)
					return false;
				m_b_name_keep = scanner.t_Cur_Token().n_Value() != 0;
			} else if(!strcmp(node_name.c_str(), "name-pattern")) {
				if(scanner.n_Cur_Token() != CMaterialLexer::token_String)
					return false;
				m_s_namepattern = scanner.t_Cur_Token().p_s_Value();
			} else if(!strcmp(node_name.c_str(), "max-width")) {
				if(scanner.n_Cur_Token() != CMaterialLexer::token_Int)
					return false;
				m_n_width = scanner.t_Cur_Token().n_Value();
			} else if(!strcmp(node_name.c_str(), "max-height")) {
				if(scanner.n_Cur_Token() != CMaterialLexer::token_Int)
					return false;
				m_n_height = scanner.t_Cur_Token().n_Value();
			} else if(!strcmp(node_name.c_str(), "handle-rotated")) {
				if(scanner.n_Cur_Token() != CMaterialLexer::token_Bool)
					return false;
				m_b_handle_rotated = scanner.t_Cur_Token().n_Value() != 0;
			} else if(!strcmp(node_name.c_str(), "histogram-enable")) {
				if(scanner.n_Cur_Token() != CMaterialLexer::token_Bool)
					return false;
				m_b_hist_eq = scanner.t_Cur_Token().n_Value() != 0;
			} else if(!strcmp(node_name.c_str(), "histogram-conditional")) {
				if(scanner.n_Cur_Token() != CMaterialLexer::token_Bool)
					return false;
				m_b_conditional_hist = scanner.t_Cur_Token().n_Value() != 0;
			} else if(!strcmp(node_name.c_str(), "max-feetspace")) {
				if(scanner.n_Cur_Token() != CMaterialLexer::token_Int)
					return false;
				m_n_below = scanner.t_Cur_Token().n_Value();
			} else if(!strcmp(node_name.c_str(), "max-headspace")) {
				if(scanner.n_Cur_Token() != CMaterialLexer::token_Int)
					return false;
				m_n_above = scanner.t_Cur_Token().n_Value();
			} else if(!strcmp(node_name.c_str(), "enhance-contrast-only")) {
				if(scanner.n_Cur_Token() != CMaterialLexer::token_Bool)
					return false;
				m_b_contrast_only = scanner.t_Cur_Token().n_Value() != 0;
			} else if(!strcmp(node_name.c_str(), "max-thumb-width")) {
				if(scanner.n_Cur_Token() != CMaterialLexer::token_Int)
					return false;
				m_n_thumb_width = scanner.t_Cur_Token().n_Value();
			} else if(!strcmp(node_name.c_str(), "max-thumb-height")) {
				if(scanner.n_Cur_Token() != CMaterialLexer::token_Int)
					return false;
				m_n_thumb_height = scanner.t_Cur_Token().n_Value();
			} else if(!strcmp(node_name.c_str(), "html-template")) {
				if(scanner.n_Cur_Token() != CMaterialLexer::token_CData)
					return false;
				m_s_template = scanner.t_Cur_Token().p_s_Value();
			} else if(!strcmp(node_name.c_str(), "grid-width")) {
				if(scanner.n_Cur_Token() != CMaterialLexer::token_Int)
					return false;
				m_n_grid_w = scanner.t_Cur_Token().n_Value();
			} else if(!strcmp(node_name.c_str(), "grid-height")) {
				if(scanner.n_Cur_Token() != CMaterialLexer::token_Int)
					return false;
				m_n_grid_h = scanner.t_Cur_Token().n_Value();
			} else if(!strcmp(node_name.c_str(), "style-empty-cell")) {
				if(scanner.n_Cur_Token() != CMaterialLexer::token_Bool)
					return false;
				m_b_empty_set_style = scanner.t_Cur_Token().n_Value() != 0;
			} else if(!strcmp(node_name.c_str(), "empty-style-name")) {
				if(scanner.n_Cur_Token() != CMaterialLexer::token_String)
					return false;
				m_s_empty_style = scanner.t_Cur_Token().p_s_Value();
			} else if(!strcmp(node_name.c_str(), "output-file")) {
				if(scanner.n_Cur_Token() != CMaterialLexer::token_String)
					return false;
				m_s_out_html = scanner.t_Cur_Token().p_s_Value();
			} else if(!strcmp(node_name.c_str(), "skip-images")) {
				if(scanner.n_Cur_Token() != CMaterialLexer::token_Bool)
					return false;
				m_b_have_files = scanner.t_Cur_Token().n_Value() != 0;
			}

			  else if(!strcmp(node_name.c_str(), "to-lowcase")) {
				if(scanner.n_Cur_Token() != CMaterialLexer::token_Bool)
					return false;
				m_b_tolower = scanner.t_Cur_Token().n_Value() != 0;
			} else if(!strcmp(node_name.c_str(), "underline-conversion")) {
				if(scanner.n_Cur_Token() != CMaterialLexer::token_Bool)
					return false;
				m_b_underline_conv = scanner.t_Cur_Token().n_Value() != 0;
			} else if(!strcmp(node_name.c_str(), "use-xhtml")) {
				if(scanner.n_Cur_Token() != CMaterialLexer::token_Bool)
					return false;
				m_b_xhtml = scanner.t_Cur_Token().n_Value() != 0;
			} else if(!strcmp(node_name.c_str(), "target-blank")) {
				if(scanner.n_Cur_Token() != CMaterialLexer::token_Bool)
					return false;
				m_b_target_blank = scanner.t_Cur_Token().n_Value() != 0;
			} else if(!strcmp(node_name.c_str(), "auto-mkdir")) {
				if(scanner.n_Cur_Token() != CMaterialLexer::token_Bool)
					return false;
				m_b_auto_mkdir = scanner.t_Cur_Token().n_Value() != 0;
			} else if(!strcmp(node_name.c_str(), "confirm-overwrite")) {
				if(scanner.n_Cur_Token() != CMaterialLexer::token_Bool)
					return false;
				m_b_overwrite_confirm = scanner.t_Cur_Token().n_Value() != 0;
			} else if(!strcmp(node_name.c_str(), "to-subfolder")) {
				if(scanner.n_Cur_Token() != CMaterialLexer::token_Bool)
					return false;
				m_b_to_subfolder = scanner.t_Cur_Token().n_Value() != 0;
			} else if(!strcmp(node_name.c_str(), "subfolder-name")) {
				if(scanner.n_Cur_Token() != CMaterialLexer::token_String)
					return false;
				m_s_subfolder = scanner.t_Cur_Token().p_s_Value();
			}
			// v1.1 new features

			if(!scanner.Get_NextToken(CMaterialLexer::token_Semicolon))
				return false;
		}

		return true;
	}

	bool LoadConfig()
	{
		CMaterialLexer scanner("config.ini");
		if(!scanner.b_Status())
			return true;

		if(!_Parse(scanner)) {
			char p_s_mess[256];
			sprintf(p_s_mess, "Error in \'config.ini\' on line %d, column %d",
				scanner.n_Cur_Line() + 1, scanner.n_Cur_Column() + 1);
			MessageBox(m_h_dlg, p_s_mess, "GalleryMaker", MB_OK);
			return false;
		}

		Update_Dialog();
		// update dialog

		return true;
	}

	void EnableWindows(bool b_enable) const
	{
		EnableWindow(GetDlgItem(m_h_dlg, IDC_INFOLDER), b_enable);
		EnableWindow(GetDlgItem(m_h_dlg, IDC_OUTFOLDER), b_enable);
		EnableWindow(GetDlgItem(m_h_dlg, IDC_NAME_KEEP), b_enable);
		EnableWindow(GetDlgItem(m_h_dlg, IDC_NAME_RENAME), b_enable);
		EnableWindow(GetDlgItem(m_h_dlg, IDC_NAMEPATTERN), b_enable);
		EnableWindow(GetDlgItem(m_h_dlg, IDC_WIDTH), b_enable);
		EnableWindow(GetDlgItem(m_h_dlg, IDC_HEIGHT), b_enable);
		EnableWindow(GetDlgItem(m_h_dlg, IDC_HANDLE_ROTATED), b_enable);
		EnableWindow(GetDlgItem(m_h_dlg, IDC_HISTO_EQ), b_enable);
		EnableWindow(GetDlgItem(m_h_dlg, IDC_CONDITIONAL_HIST), b_enable);
		EnableWindow(GetDlgItem(m_h_dlg, IDC_BELOW), b_enable);
		EnableWindow(GetDlgItem(m_h_dlg, IDC_ABOVE), b_enable);
		EnableWindow(GetDlgItem(m_h_dlg, IDC_CONTRAST_ONLY), b_enable);
		EnableWindow(GetDlgItem(m_h_dlg, IDC_COLORS), b_enable);
		EnableWindow(GetDlgItem(m_h_dlg, IDC_THUMB_WIDTH), b_enable);
		EnableWindow(GetDlgItem(m_h_dlg, IDC_THUMB_HEIGHT), b_enable);
		EnableWindow(GetDlgItem(m_h_dlg, IDC_TEMPLATE), b_enable);
		EnableWindow(GetDlgItem(m_h_dlg, IDC_GRID_W), b_enable);
		EnableWindow(GetDlgItem(m_h_dlg, IDC_GRID_H), b_enable);
		EnableWindow(GetDlgItem(m_h_dlg, IDC_EMPTY_SET_STYLE), b_enable);
		EnableWindow(GetDlgItem(m_h_dlg, IDC_EMPTY_STYLE), b_enable);
		EnableWindow(GetDlgItem(m_h_dlg, IDC_OUT_HTML), b_enable);
		EnableWindow(GetDlgItem(m_h_dlg, IDC_SKIP_IMAGES), b_enable);
		EnableWindow(GetDlgItem(m_h_dlg, IDOK), b_enable);
		EnableWindow(GetDlgItem(m_h_dlg, IDCANCEL), b_enable);
		EnableWindow(GetDlgItem(m_h_dlg, IDC_DEFAULTS), b_enable); // !!

		EnableWindow(GetDlgItem(m_h_dlg, IDC_TOLOWER), b_enable);
		EnableWindow(GetDlgItem(m_h_dlg, IDC_UNDERLINE_CONV), b_enable);
		EnableWindow(GetDlgItem(m_h_dlg, IDC_XHTML), b_enable);
		EnableWindow(GetDlgItem(m_h_dlg, IDC_TARGET_BLANK), b_enable);
		EnableWindow(GetDlgItem(m_h_dlg, IDC_AUTO_MKDIR), b_enable);
		EnableWindow(GetDlgItem(m_h_dlg, IDC_OVERWRITE_CONFIRM), b_enable);
		EnableWindow(GetDlgItem(m_h_dlg, IDC_BREAK), !b_enable); // enable if others are disabled and vice versa
		EnableWindow(GetDlgItem(m_h_dlg, IDC_SUBFOLDER), b_enable);
		EnableWindow(GetDlgItem(m_h_dlg, IDC_SUBFOLDER_NAME), b_enable);
		// v1.1 new features
	}

	class Pixel3f {
	public:
		float p_data[3];

		inline Pixel3f()
		{
			for(int n = 0; n < 3; ++ n)
				p_data[n] = 0;
		}

		inline Pixel3f(const Pixel3f &r_pixel)
		{
			for(int n = 0; n < 3; ++ n)
				p_data[n] = r_pixel.p_data[n];
		}

		inline Pixel3f(unsigned __int32 n_pixel)
		{
			for(int n = 0; n < 3; ++ n)
				p_data[n] = (n_pixel >> (8 * n)) & 0xff;
		}

		inline const Pixel3f &operator *=(float f_scale)
		{
			for(int n = 0; n < 3; ++ n)
				p_data[n] *= f_scale;
			return *this;
		}

		inline const Pixel3f &operator +=(const Pixel3f &r_pixel)
		{
			for(int n = 0; n < 3; ++ n)
				p_data[n] += r_pixel.p_data[n];
			return *this;
		}

		inline Pixel3f operator *(float f_scale) const
		{
			Pixel3f pix(*this);
			for(int n = 0; n < 3; ++ n)
				pix.p_data[n] *= f_scale;
			return pix;
		}

		inline Pixel3f operator +(const Pixel3f &r_pixel) const
		{
			Pixel3f pix(*this);
			for(int n = 0; n < 3; ++ n)
				pix.p_data[n] += r_pixel.p_data[n];
			return pix;
		}

		inline unsigned __int32 n_NotNegative_Convert() const
		{
			int p_int_data[3] = {p_data[0], p_data[1], p_data[2]};
			unsigned __int32 n_result = 0;
			for(int n = 0; n < 3; ++ n) {
				int n_channel = p_int_data[n];
				n_channel = (n_channel < 255)? n_channel : 255; // clamp data
				n_result += n_channel << (8 * n);
			}
			return n_result;
		}

		inline unsigned __int32 n_UnClamped_Convert() const
		{
			int p_int_data[3] = {p_data[0], p_data[1], p_data[2]};
			unsigned __int32 n_result = 0;
			for(int n = 0; n < 3; ++ n)
				n_result += (int)p_int_data[n] << (8 * n);
			return n_result;
		}

		inline operator unsigned __int32() const
		{
			int p_int_data[3] = {p_data[0], p_data[1], p_data[2]};
			unsigned __int32 n_result = 0;
			for(int n = 0; n < 3; ++ n) {
				int n_channel = p_int_data[n];
				n_channel = (n_channel < 255)? ((n_channel >= 0)? n_channel : 0) : 255; // clamp data
				n_result += n_channel << (8 * n);
			}
			return n_result;
		}
	};

	class CPixelRGB32 {
	protected:
		unsigned __int32 m_n_value;

	public:
		CPixelRGB32(unsigned __int32 n_value)
			:m_n_value(n_value)
		{ }

		inline operator unsigned __int32() const
		{
			return m_n_value;
		}

		inline CPixelRGB32 operator *(float f_multiplier) const
		{
			_ASSERTE(f_multiplier >= 0 && f_multiplier <= 1);
			unsigned short n_fp_multiplier = (unsigned short)(0xff * f_multiplier);
			return CPixelRGB32(((((m_n_value & 0xff00ff) * n_fp_multiplier) & 0xff00ff00) |
				(((m_n_value & 0x00ff00) * n_fp_multiplier) & 0x00ff0000)) >> 8);
		}
	};

	static unsigned __int32 n_2D_Sampler(const TBmp *p_image, float f_x, float f_y)
	{
		int n_width = p_image->n_width;
		int n_height = p_image->n_height;
		int n_x = (int)f_x;
		int n_y = n_width * (int)f_y;
		int n_x1 = (n_x + 1 >= n_width)? n_width - 1 : n_x + 1;
		int n_y1 = ((n_y + 1 >= n_height * n_width)? (n_height - 1) * n_width : n_y + n_width);

		float f_dx = f_x - n_x;
		float f_dy = f_y - (int)f_y;
		float f_dx1 = 1 - f_dx;
		float f_dy1 = 1 - f_dy;

		return (CPixelRGB32(p_image->p_buffer[n_x1 + n_y1]) * (f_dx * f_dy)) +
			(CPixelRGB32(p_image->p_buffer[n_x + n_y1]) * (f_dx1 * f_dy)) +
			(CPixelRGB32(p_image->p_buffer[n_x1 + n_y]) * (f_dx * f_dy1)) +
			(CPixelRGB32(p_image->p_buffer[n_x + n_y]) * (f_dx1 * f_dy1));
	}

	static Pixel3f v_2D_Sampler(const TBmp *p_image,
		float f_coord1_x, float f_coord1_y, float f_coord2_x, float f_coord2_y)
	{
		Pixel3f v_color;
		float f_a_tot = 0;
		for(float f_y = f_coord1_y; f_y < f_coord2_y;) {
			// prev y

			float f_dy;
			if(f_y == f_coord1_y && f_y - floorf(f_y) > 1e-3) {
				float f_py = f_y;
				f_y = (ceilf(f_y) < f_coord2_y)? ceilf(f_y) : f_coord2_y;
				f_dy = f_y - f_py;
			} else if(f_coord2_y - f_y < 1) {
				f_dy = f_coord2_y - f_y;
				f_y = f_coord2_y;
			} else {
				f_y += f_dy = 1;
			}

			if(f_dy > 1e-3f) {
				for(float f_x = f_coord1_x; f_x < f_coord2_x;) {
					float f_a;
					if(f_x == f_coord1_x && f_x - floorf(f_x) > 1e-3) {
						float f_px = f_x;
						f_x = (ceilf(f_x) < f_coord2_x)? ceilf(f_x) : f_coord2_x;
						f_a = f_dy * (f_x - f_px);
					} else if(f_coord2_x - f_x < 1) {
						f_a = f_dy * (f_coord2_x - f_x);
						f_x = f_coord2_x;
					} else {
						f_x += 1;
						f_a = f_dy;
					}

					f_a_tot += f_a;
					v_color += Pixel3f(p_image->p_buffer[(int)f_x + (int)f_y * p_image->n_width]
						/*n_2D_Sampler(p_image, f_px + (f_x - f_px) * .5f, f_py + f_dy * .5f)*/) * f_a;
				}
			}
		}

		return v_color * (1.0f / f_a_tot);
	}

	static TBmp *p_Image_ResizeMinify(const TBmp *p_image, int n_new_width, int n_new_height)
	{
		TBmp *p_new_image;
		if(!(p_new_image = new TBmp))
			return false;
		if(!(p_new_image->p_buffer = new unsigned __int32[n_new_width * n_new_height])) {
			delete p_new_image;
			return false;
		}
		p_new_image->n_width = n_new_width;
		p_new_image->n_height = n_new_height;
		p_new_image->b_grayscale = p_image->b_grayscale;
		p_new_image->b_alpha = p_image->b_alpha;
		p_new_image->n_former_bpp = p_image->n_former_bpp;
		// init new bitmap

		float f_dx = (float)p_image->n_width / (n_new_width + 1);
		float f_dy = (float)p_image->n_height / (n_new_height + 1), f_y = 0;
		unsigned __int32 *p_out = p_new_image->p_buffer;
		for(int j = 0; j < n_new_height; ++ j, f_y += f_dy) {
			float f_x = 0;
			for(int i = 0; i < n_new_width; ++ i, f_x += f_dx, ++ p_out) {
				*p_out = v_2D_Sampler(p_image, f_x, f_y,
					f_x + f_dx, f_y + f_dy).n_UnClamped_Convert();
			}
		}
		// resample bitmap

		return p_new_image;
	}

	static bool CalcSize(int &n_width, int &n_height,
		int n_max_width, int n_max_height, bool b_pixel_count_only)
	{
		if(b_pixel_count_only) {
			float f_max_pixels = n_max_width * n_max_height;
			float f_pixels = n_width * n_height;
			if(f_pixels > f_max_pixels) {
				float f_scale = (float)sqrt(f_max_pixels / f_pixels);
				n_width *= f_scale;
				n_height *= f_scale;
				return true; // resize
			} else
				return false; // do not resize
		} else {
			if(n_width <= n_max_width && n_height <= n_max_height)
				return false; // no need to resize

			float f_w_ratio = (float)n_max_width / n_width;
			if((int)(n_height * f_w_ratio) <= n_max_height) {
				float f_h_ratio = (float)n_max_height / n_height;
				if(n_width * (int)(n_height * f_w_ratio) >=
				   (int)(n_width * f_h_ratio) * n_max_height) {
					n_width = n_max_width;
					n_height = (int)(n_height * f_w_ratio);
				} else {
					n_width = (int)(n_width * f_h_ratio);
					n_height = n_max_height;
				}
				// can still scale to fit height ... in case picture doesn't
				// have the same side ratio as max dimensions, it could be even better
			} else {
				float f_h_ratio = (float)n_max_height / n_height;
				n_width = (int)(n_width * f_h_ratio);
				n_height = n_max_height;
			}
			_ASSERTE(n_width <= n_max_width && n_height <= n_max_height);
			return true; // resize
		}
	}

	void DrawThumbnail(bool b_left, const TBmp *p_bitmap)
	{
		CGDI_Bitmap *p_dib = (b_left)? m_p_left : m_p_right;
		if(!p_dib || !p_dib->b_Status())
			return;

		int n_width = p_dib->n_Width();
		int n_height = p_dib->n_Height();
		if(p_bitmap) {
			memset(&(*p_dib)[0], 0, n_width * n_height * sizeof(unsigned __int32));
			// clear background

			int n_w = p_bitmap->n_width, n_h = p_bitmap->n_height;
			CalcSize(n_w, n_h, n_width, n_height, false);
			int n_off_x = (n_width - n_w) / 2;
			int n_off_y = (n_height - n_h) / 2;
			int n_scale_x = p_bitmap->n_width / n_w;
			int n_scale_y = p_bitmap->n_height / n_h;
			n_scale_y *= p_bitmap->n_width;
			// calculate scaling factors

			const unsigned __int32 *p_buffer = p_bitmap->p_buffer;
			unsigned __int32 *p_out = &(*p_dib)[0] + n_off_x + n_off_y * n_width;
			for(int y = 0, n_more_shift = n_width - n_w; y < n_h;
			   ++ y, p_buffer += n_scale_y, p_out += n_more_shift) {
				const unsigned __int32 *p_scanline = p_buffer;
				for(int x = 0; x < n_w; ++ x, p_scanline += n_scale_x, ++ p_out)
					*p_out = n_RGB_to_BGR(*p_scanline);
			}
			// simple nearest neighbour
		}

		HWND h_wnd = GetDlgItem(m_h_dlg, (b_left)? IDC_BEFORE : IDC_AFTER);
		HDC h_dc = GetDC(h_wnd);
		p_dib->DrawBitmap(h_dc, 0, n_height - 1, n_width, -n_height);
		ReleaseDC(h_wnd, h_dc);
		// display using GDI
	}

	static inline unsigned __int32 n_RGB_to_BGR(unsigned __int32 n_pixel)
	{
		return (n_pixel & 0xff00) | ((n_pixel & 0xff) << 16) | ((n_pixel >> 16) & 0xff);
	}

	static inline unsigned __int32 n_Grey(unsigned __int32 n_color)
	{
		const unsigned short n_red_weight = (unsigned short)(.299f * 0x10000);
		const unsigned short n_green_weight = (unsigned short)(.587f * 0x10000);
		const unsigned short n_blue_weight = (unsigned short)(.114f * 0x10000);
		// 0:16 fixed-point PAL weights
		
		unsigned char n_grey =
			(unsigned char)((n_red_weight * ((n_color >> 16) & 0xff) +
			n_green_weight * ((n_color >> 8) & 0xff) +
			n_blue_weight * (n_color & 0xff) + 0x8000) >> 16);
		// integer arithmetic intensity calculation
		// uint because result is multiplication of 16 * 8 bits ~ 24 bits max
		// 0x8000 ~ .5 in 0:16 FP

		return n_grey | (n_grey << 8) | (n_grey << 16);
	}

	bool RenameFile(int n_index, std::string &r_s_filename)
	{
		if(!m_b_name_keep) {
			char *p_s_new_filename = p_s_Format(m_s_namepattern.c_str(), n_index);
			if(!p_s_new_filename)
				return false;
			r_s_filename = p_s_new_filename;
			delete[] p_s_new_filename;
		} else {
			if(m_b_tolower) {
				for(int i = 0; i < r_s_filename.length(); ++ i) {
					if(r_s_filename[i] >= 'A' && r_s_filename[i] <= 'Z')
						r_s_filename[i] = tolower(r_s_filename[i]);
				}
			}
			// make all leters lowcase (won't work for accents)

			if(m_b_underline_conv) {
				for(int i = 0; i < r_s_filename.length(); ++ i) {
					if(r_s_filename[i] == ' ')
						r_s_filename[i] = '_';
				}
			}
			// convert spaces to underlines
		}
		// create output filename
		
		if(m_b_to_subfolder)
			r_s_filename = m_s_subfolder + r_s_filename;
		// move files to subfolder

		return true;
	}

	static void CreatePath(const char *p_s_path)
	{
		char p_s_dir_name[1024];
		const char *p_path_ptr = p_s_path;
		char *p_dirname_ptr = p_s_dir_name; 
		while(*p_path_ptr) {
			if(('\\' == *p_path_ptr) || ('/' == *p_path_ptr)) {
				if(':' != *(p_path_ptr - 1))
					CreateDirectory(p_s_dir_name, NULL);
			}
			*p_dirname_ptr ++ = *p_path_ptr ++;
			*p_dirname_ptr = '\0';
		}
		//CreateDirectory(p_s_dir_name, NULL); // this is filename
	}

	bool ProcessFile(int n_index, std::string &r_s_filename)
	{
		FILE *p_fr;
		if(!(p_fr = fopen((m_s_infolder + r_s_filename).c_str(), "rb"))) {
			MessageBox(m_h_dlg, ("Error: error opening \'" + m_s_infolder +
				r_s_filename + "\' (file not found)").c_str(), "GalleryMaker", MB_OK);
			return false;
		}
		CJpegDecoder decoder;
		TBmp *p_bitmap = decoder.p_Decode_JPEG(p_fr);
		fclose(p_fr);
		if(!p_bitmap) {
			MessageBox(m_h_dlg, ("Error: error decoding \'" + m_s_infolder +
				r_s_filename + "\' (invalid jpeg file)").c_str(), "GalleryMaker", MB_OK);
			return false;
		}
		// attempt to load jpeg from file ...

		for(unsigned __int32 *p_pixel = p_bitmap->p_buffer,
		   *p_end = p_bitmap->p_buffer + (p_bitmap->n_width * p_bitmap->n_height);
		   p_pixel != p_end; ++ p_pixel)
			*p_pixel = n_RGB_to_BGR(*p_pixel);
		// bgr <-> rgb

		DrawThumbnail(true, p_bitmap);
		// draw left thumbnail

		int n_w = p_bitmap->n_width, n_h = p_bitmap->n_height;
		if(CalcSize(n_w, n_h, m_n_width, m_n_height, m_b_handle_rotated)) {
			TBmp *p_resized;
			if(!(p_resized = p_Image_ResizeMinify(p_bitmap, n_w, n_h))) {
				delete[] p_bitmap->p_buffer;
				delete p_bitmap;
				MessageBox(m_h_dlg, "Error: not enough memory", "GalleryMaker", MB_OK);
				return false;
			}
			delete[] p_bitmap->p_buffer;
			delete p_bitmap;
			p_bitmap = p_resized;
		}
		// resize input if neccesary

		if(m_b_hist_eq) {
			int p_count[3 * 256] = {0}; // can this lie on stack?
			int *p_red_ptr = p_count;
			int *p_green_ptr = p_count + 256;
			int *p_blue_ptr = p_count + 512;
			// histogram tables

			if(!m_b_contrast_only) {
				for(const unsigned __int32 *p_pixel = p_bitmap->p_buffer,
				   *p_end = p_bitmap->p_buffer + (p_bitmap->n_width * p_bitmap->n_height);
				   p_pixel != p_end; ++ p_pixel) {
					++ p_red_ptr[(*p_pixel >> 16) & 0xff];
					++ p_green_ptr[(*p_pixel >> 8) & 0xff];
					++ p_blue_ptr[*p_pixel & 0xff];
				}
				// calculate color histogram
			} else {
				for(const unsigned __int32 *p_pixel = p_bitmap->p_buffer,
				   *p_end = p_bitmap->p_buffer + (p_bitmap->n_width * p_bitmap->n_height);
				   p_pixel != p_end; ++ p_pixel)
					++ p_red_ptr[n_Grey(*p_pixel) & 0xff];
				// calculate greyscale histogram
			}

			unsigned char p_min[3], p_max[3];
			int n_tolerance = (int)((p_bitmap->n_width * p_bitmap->n_height) * .001f);
			for(int i = 0, n = (m_b_contrast_only)? 1 : 3; i < n; ++ i) {
				const int *p_histogram = p_count + (256 * i);

				int n_min = 0;
				for(int n_space = n_tolerance - p_histogram[0]; n_space >= 0 && n_min < 256 - 1;)
					n_space -= p_histogram[++ n_min];
				p_min[i] = n_min;
				// determine min cutoff

				int n_max = 255;
				for(int n_space = n_tolerance - p_histogram[255]; n_space >= 0 && n_max > n_min + 1;)
					n_space -= p_histogram[-- n_max];
				p_max[i] = n_max;
				// determine max cutoff
			}
			if(m_b_contrast_only) {
				for(int i = 1; i < 3; ++ i) {
					p_min[i] = p_min[0];
					p_max[i] = p_max[0];
				}
				// all channels will scale the same
			}
			// determine min, max, with some tolerance to random spikes

			if(!m_b_conditional_hist ||
			   p_min[0] > m_n_above || p_min[1] > m_n_above || p_min[2] > m_n_above ||
			   p_max[0] < m_n_below || p_max[1] < m_n_below || p_max[2] < m_n_below) {
				int n_red_subtract = p_min[0] << 8; // xxx
				int n_green_subtract = p_min[1] << 8;
				int n_blue_subtract = p_min[2];
				// save some shifting work

				unsigned short n_red_scale = 0x10000 / (p_max[0] - p_min[0]);
				unsigned short n_green_scale = 0x10000 / (p_max[1] - p_min[1]);
				unsigned short n_blue_scale = 0x10000 / (p_max[2] - p_min[2]);
				// scales in 8:8 fixed point

				for(unsigned __int32 *p_pixel = p_bitmap->p_buffer,
				   *p_end = p_bitmap->p_buffer + (p_bitmap->n_width * p_bitmap->n_height);
				   p_pixel != p_end; ++ p_pixel) {
					int n_red = (*p_pixel & 0xff0000) >> 8; // xxx (have to shift red down otherwise it could overflow)
					int n_green = *p_pixel & 0xff00;
					int n_blue = *p_pixel & 0xff;
					// get pixel colors

					n_red -= n_red_subtract;
					n_green -= n_green_subtract;
					n_blue -= n_blue_subtract;
					// subtract minimum values, clamp

					n_red *= n_red_scale; // xxx
					n_green *= n_green_scale;
					n_blue *= n_blue_scale;
					// scale by fp8:8 values

					if(n_red < 0) // xxx
						n_red = 0;
					else if(n_red > 0xff0000)
						n_red = 0xff0000;
					else
						n_red = n_red & 0xff0000;
					if(n_green < 0)
						n_green = 0;
					else if(n_green > 0xff0000)
						n_green = 0xff0000;
					else
						n_green = n_green & 0xff0000;
					if(n_blue < 0)
						n_blue = 0;
					else if(n_blue > 0xff00)
						n_blue = 0xff00;
					else
						n_blue = n_blue & 0xff00;
					// debug this!

					*p_pixel = n_red | ((n_green | n_blue) >> 8); // xxx
					// pack values
				}
				// color histogram equalization
			}
			// shall we proceed with histogram equalization?
		}
		// do histogram equalization

		DrawThumbnail(false, p_bitmap);
		// draw right thumbnail

		CJpegEncoder encoder;
		encoder.Write_QuantTable(0, true);
		encoder.Write_QuantTable(1, true);
		unsigned char p_quant_tab[64];
		CQuantTableFactory::Calc_QuantTable(p_quant_tab, true, 1.0f);
		encoder.Set_QuantTable(0, p_quant_tab);
		CQuantTableFactory::Calc_QuantTable(p_quant_tab, false, 1.0f);
		encoder.Set_QuantTable(1, p_quant_tab);
		// config jpeg encoder

		if(!RenameFile(n_index, r_s_filename)) {
			delete[] p_bitmap->p_buffer;
			delete p_bitmap;
			MessageBox(m_h_dlg, "Error: not enough memory", "GalleryMaker", MB_OK);
			return false;
		}
		// create output filename

		if(m_b_auto_mkdir)
			CreatePath((m_s_outfolder + r_s_filename).c_str());
		// create path

		do {
			if(!_access((m_s_outfolder + r_s_filename).c_str(), 0)) {
				if(m_b_overwrite_confirm) {
					int n_result;
					if((n_result = MessageBox(m_h_dlg, ("File \'" +
					   m_s_outfolder + r_s_filename + "\' exists. Overwrite?").c_str(),
					   "GalleryMaker", MB_YESNOCANCEL)) == IDNO) {
						break;
					} else if(n_result == IDCANCEL) {
						m_b_abort = true;
						return false;
					}
				}
			}
			// does the file exist?

			bool b_result = false;
			FILE *p_fw;
			if(p_fw = fopen((m_s_outfolder + r_s_filename).c_str(), "wb")) {
				b_result = encoder.Encode_Jpeg(p_fw, p_bitmap);
				fclose(p_fw);
			}
			if(!b_result) {
				delete[] p_bitmap->p_buffer;
				delete p_bitmap;
				MessageBox(m_h_dlg, ("Error: error writing \'" + m_s_outfolder +
					r_s_filename + "\'").c_str(), "GalleryMaker", MB_OK);
				return false;
			}
		} while(0);
		// write image

		int n_w2 = p_bitmap->n_width, n_h2 = p_bitmap->n_height;
		if(CalcSize(n_w2, n_h2, m_n_thumb_width, m_n_thumb_height, false)) {
			TBmp *p_resized;
			if(!(p_resized = p_Image_ResizeMinify(p_bitmap, n_w2, n_h2))) {
				delete[] p_bitmap->p_buffer;
				delete p_bitmap;
				return false;
				MessageBox(m_h_dlg, "Error: not enough memory", "GalleryMaker", MB_OK);
			}
			delete[] p_bitmap->p_buffer;
			delete p_bitmap;
			p_bitmap = p_resized;
		}
		// resize to thumbnail

		do {
			std::string s_thumb_filename = r_s_filename.substr(0,
			   r_s_filename.length() - 4) + "_thumb.jpg";
			if(!_access((m_s_outfolder + s_thumb_filename).c_str(), 0)) {
				if(m_b_overwrite_confirm) {
					int n_result;
					if((n_result = MessageBox(m_h_dlg, ("File \'" +
					   m_s_outfolder + s_thumb_filename + "\' exists. Overwrite?").c_str(),
					   "GalleryMaker", MB_YESNOCANCEL)) == IDNO) {
						break;
					} else if(n_result == IDCANCEL) {
						m_b_abort = true;
						return false;
					}
				}
			}
			// does the file exist?

			bool b_result = false;
			FILE *p_fw;
			if(p_fw = fopen((m_s_outfolder + s_thumb_filename).c_str(), "wb")) {
				b_result = encoder.Encode_Jpeg(p_fw, p_bitmap);
				fclose(p_fw);
			}
			if(!b_result) {
				delete[] p_bitmap->p_buffer;
				delete p_bitmap;
				MessageBox(m_h_dlg, ("Error: error writing \'" + m_s_outfolder +
					s_thumb_filename + "\'").c_str(), "GalleryMaker", MB_OK);
				return false;
			}
		} while(0);
		// write image thumbnail

		delete[] p_bitmap->p_buffer;
		delete p_bitmap;
		// cleanup

		return true;
	}

	bool ReplaceAll(std::string &r_s_string, const char *p_s_pattern, const char *p_s_replacement)
	{
		int n_pattern_len = strlen(p_s_pattern);
		int n_replacement_len = strlen(p_s_replacement);
		int n_pos;
		while((n_pos = r_s_string.find(p_s_pattern)) != std::string::npos) {
			r_s_string.erase(r_s_string.begin() + n_pos, r_s_string.begin() + n_pos + n_pattern_len);
			// vyjme vzor

			if(r_s_string.capacity() < r_s_string.length() + n_replacement_len) {
				int n_reserve = (r_s_string.length())? r_s_string.length() * 2 : n_replacement_len;
				while(n_reserve < r_s_string.length() + n_replacement_len)
					n_reserve *= 2;
				r_s_string.reserve(n_reserve);
				if(r_s_string.capacity() < r_s_string.length() + n_replacement_len)
					return false;
			}

			r_s_string.insert(n_pos, p_s_replacement);
			// vlozi nahradu
		}

		return true;
	}

	bool Proceed()
	{
		m_b_abort = false;

		if(!Update_Data()) {
			MessageBox(m_h_dlg, "Error: windows error", "GalleryMaker", MB_OK);
			return false;
		}

		if(m_b_target_blank && m_b_xhtml) {
			MessageBox(m_h_dlg, "Error: target=\"_blank\" and XHTML 1.0 can't be used together.\n"
				"Deselect one of them and try again.", "GalleryMaker", MB_OK);
			return false;
		}

		if(!m_s_infolder.length() || !m_s_infolder.length() ||
		   (!m_b_name_keep && !m_s_namepattern.length()) ||
		   !m_s_template.length() || !m_s_out_html.length() ||
		   (m_b_empty_set_style && !m_s_empty_style.length()) ||
		   (m_b_to_subfolder && !m_s_subfolder.length())) {
			MessageBox(m_h_dlg, "You left some form fields empty! Cannot continue",
				   "GalleryMaker", MB_OK);
			return false;
		}
		ReplaceAll(m_s_infolder, "/", "\\");
		ReplaceAll(m_s_outfolder, "/", "\\");
		ReplaceAll(m_s_subfolder, "/", "\\");
		ReplaceAll(m_s_namepattern, "/", "\\");
		if(m_s_infolder[m_s_infolder.length() - 1] != '\\')
			m_s_infolder += '\\';
		if(m_s_outfolder[m_s_outfolder.length() - 1] != '\\')
			m_s_outfolder += '\\';
		if(m_b_to_subfolder) {
			if(m_s_subfolder.length() && m_s_subfolder[m_s_subfolder.length() - 1] != '\\')
				m_s_subfolder += '\\';
			if(m_s_subfolder.length() && m_s_subfolder[0] == '\\')
				m_s_subfolder.erase(m_s_subfolder.begin(), m_s_subfolder.begin() + 1);
		}
		if(m_s_namepattern.length() && m_s_namepattern[0] == '\\')
			m_s_namepattern.erase(m_s_namepattern.begin(), m_s_namepattern.begin() + 1);
		// check form field values

		Update_Dialog();
		// put values back for user to see

		CDirectory dir(m_s_infolder.c_str());
		if(!dir.p_s_Name()) {
			MessageBox(m_h_dlg, "Error: not enough memory", "GalleryMaker", MB_OK);
			return false;
		}

		std::vector<std::string> file_list;

		TFileInfo t_file;
		if(!dir.Get_FirstFile(t_file)) {
			MessageBox(m_h_dlg, "Error: error getting file list (input folder does not exist?)", "GalleryMaker", MB_OK);
			return false;
		}
		int n_dir_string_length = m_s_infolder.length();
		while(t_file.b_valid) {
			const char *p_s_filename = t_file.p_s_Filename() + n_dir_string_length;
			// don't want path, just filename

			if(strlen(p_s_filename) >= 3 &&
			   !stricmp(p_s_filename + (strlen(p_s_filename) - 3), "jpg")) {
				if(file_list.capacity() == file_list.size()) {
					file_list.reserve((file_list.size())? file_list.size() * 2 : 1);
					if(file_list.capacity() == file_list.size()) {
						MessageBox(m_h_dlg, "Error: not enough memory", "GalleryMaker", MB_OK);
						return false;
					}
				}
				file_list.push_back(p_s_filename);
			}
			if(!dir.Get_NextFile(t_file)) {
				MessageBox(m_h_dlg, "Error: error getting file list", "GalleryMaker", MB_OK);
				return false;
			}
		}
		// get list of jpeg images

		std::sort(file_list.begin(), file_list.end());
		// sort by name

		SendMessage(GetDlgItem(m_h_dlg, IDC_PROGRESS),
			PBM_SETRANGE32, 0, file_list.size());
		for(unsigned int i = 0; i < file_list.size(); ++ i) {
			SendMessage(GetDlgItem(m_h_dlg, IDC_PROGRESS),
				PBM_SETPOS, i, 0);
			// update progress-bar

			if(!DoEvents())
				return false;

			std::string s_tmp = file_list[i];
			if((m_b_have_files && !RenameFile(i, file_list[i])) ||
			   (!m_b_have_files && !ProcessFile(i, file_list[i]))) {
				if(m_b_abort)
					return false;
				if(MessageBox(m_h_dlg, ("Error processing file \'" + s_tmp + "\'.\n"
				   "Do you want to skip this one and move on the rest?").c_str(),
				   "GalleryMaker", MB_YESNO) == IDNO)
					return false;
				else {
					file_list.erase(file_list.begin() + i, file_list.begin() + i + 1);
					-- i;
					continue;
				}
			}
			// saves image under new location and puts that filename
			// back to list so it will contain strings of converted images

			std::string &r_s_name = file_list[i];
			for(int i = 0; i < r_s_name.length(); ++ i) {
				if(r_s_name[i] == '\\')
					r_s_name[i] = '/';
			}
		}
		SendMessage(GetDlgItem(m_h_dlg, IDC_PROGRESS),
			PBM_SETPOS, file_list.size(), 0);
		if(!DoEvents())
			return false;
		// does the image work

		int n_file_num = file_list.size();
		int n_line_num = (n_file_num + m_n_grid_w - 1) / m_n_grid_w;
		int n_page_num = (n_line_num + m_n_grid_h - 1) / m_n_grid_h;
		// determine number of gallery pages

		int n_cur_image = 0;
		for(int i = 0; i < n_page_num; ++ i) {
			char p_s_outfile[1024];
			sprintf(p_s_outfile, m_s_out_html.c_str(), i);

			if(!_access((m_s_outfolder + p_s_outfile).c_str(), 0)) {
				if(m_b_overwrite_confirm) {
					int n_result;
					if((n_result = MessageBox(m_h_dlg, ("Gallery page \'" +
					   m_s_outfolder + p_s_outfile + "\' exists. Overwrite?").c_str(),
					   "GalleryMaker", MB_YESNOCANCEL)) == IDNO) {
						continue;
					} else if(n_result == IDCANCEL) {
						m_b_abort = true;
						return false;
					}
				}
			}
			// does the file exist?

			FILE *p_fw = fopen((m_s_outfolder + p_s_outfile).c_str(), "wb");
			if(!p_fw) {
				MessageBox(m_h_dlg, ("Error writing gallery page \'" +
					m_s_outfolder + p_s_outfile + "\'").c_str(), "GalleryMaker", MB_OK);
				return false;
			}

			std::string s_page_source = m_s_template;
			char p_s_buffer[1024];

			sprintf(p_s_buffer, "%d", i + 1);
			ReplaceAll(s_page_source, "$page-no", p_s_buffer);

			sprintf(p_s_buffer, "%d", n_page_num);
			ReplaceAll(s_page_source, "$page-num", p_s_buffer);

			if(i) {
				sprintf(p_s_buffer, m_s_out_html.c_str(), 0);
				ReplaceAll(s_page_source, "$first-file", p_s_buffer);
			} else
				ReplaceAll(s_page_source, "$first-file", "#nogo");

			if(i) {
				sprintf(p_s_buffer, m_s_out_html.c_str(), i - 1);
				ReplaceAll(s_page_source, "$prev-file", p_s_buffer);
			} else
				ReplaceAll(s_page_source, "$prev-file", "#nogo");

			if(i + 1 != n_page_num) {
				sprintf(p_s_buffer, m_s_out_html.c_str(), i + 1);
				ReplaceAll(s_page_source, "$next-file", p_s_buffer);
			} else
				ReplaceAll(s_page_source, "$next-file", "#nogo");

			if(i + 1 != n_page_num) {
				sprintf(p_s_buffer, m_s_out_html.c_str(), n_page_num - 1);
				ReplaceAll(s_page_source, "$last-file", p_s_buffer);
			} else
				ReplaceAll(s_page_source, "$last-file", "#nogo");

			std::string s_table_body;
			for(int j = 0; j < m_n_grid_h && n_cur_image < n_file_num; ++ j) {
				s_table_body += "<tr>\r\n";
				for(int i = 0; i < m_n_grid_w; ++ i, ++ n_cur_image) {
					if(n_cur_image < n_file_num) {
						const std::string &r_s_filename = file_list[n_cur_image];
						std::string s_thumb_filename = r_s_filename.substr(0, r_s_filename.length() - 4) + "_thumb.jpg";
						s_table_body += "\t<td><a href=\"" + r_s_filename +
							"\"" + ((m_b_target_blank)? " target=\"_blank\"" : "") +
							"><img src=\"" + s_thumb_filename + "\" alt=\"\"" +
							((m_b_xhtml)? "/" : "") + "></a></td>\r\n";
					} else {
						s_table_body += (m_b_empty_set_style)? ("\t<td style=\"" +
							m_s_empty_style + "\">&nbsp;</td>\r\n").c_str() : "\t<td>&nbsp;</td>\r\n";
						// empty cell, not enough images to complete it
					}
				}
				s_table_body += "</tr>\r\n";
			}

			ReplaceAll(s_page_source, "$gallery-content", s_table_body.c_str());

			if(fprintf(p_fw, "%s", s_page_source.c_str()) != s_page_source.length()) {
				fclose(p_fw);
				MessageBox(m_h_dlg, ("Insufficient free space while writing page \'" +
					m_s_outfolder + p_s_outfile + "\'").c_str(), "GalleryMaker", MB_OK);
				return false;
			}

			fclose(p_fw);
		}

		return true;
	}

	bool DoEvents()
	{
		MSG t_msg;
		for(int i = 0; i < 500; ++ i) {
			if(PeekMessage(&t_msg, m_h_dlg, 0, 0, PM_REMOVE)) {
				TranslateMessage(&t_msg);
				DispatchMessage(&t_msg);
			} else
				break;
		}
		// do events

		return !m_b_breakop;
	}
};

int WINAPI WinMain(HINSTANCE h_instance, HINSTANCE h_prev_instance, PSTR p_s_cmd_line, int n_cmd_show)
{
	InitCommonControls();
	// commctrl init

	CMainWindow wnd;
	if(!wnd.Create(0, GetModuleHandle(NULL))) {
		MessageBox(0, "Error creating dialog window."
			" Maybe common controls not supported?", "GalleryMaker", MB_OK);
		return -1;
	}
	// create window

	MSG t_msg;
	while(GetMessage(&t_msg, NULL, 0, 0) > 0 && wnd.b_Created()) {
		TranslateMessage(&t_msg);
		DispatchMessage(&t_msg);
	}
	// mess loop

	return 0;
}
