/*
								+----------------------------------+
								|                                  |
								| *** berLame material system *** |
								|                                  |
								|   Copyright  -tHE SWINe- 2007   |
								|                                  |
								|            Material.h            |
								|                                  |
								+----------------------------------+
*/

/*
 *	passed code violation at 2007-09-27
 *
 *	removed all but material lexer, using this to parse config file
 *	might be buggy (excerpted from not very well debugged code), do not copy
 *
 */

#ifndef __MATERIAL_SYSTEM_INCLUDED
#define __MATERIAL_SYSTEM_INCLUDED

/*
 *	MATERIAL_LEXER_TOKEN_DUMP
 *		- if defined, class CMaterialLexer::TToken has member
 *		  function Dump() which prints token contents to stdout
 */
#define MATERIAL_LEXER_TOKEN_DUMP

/*
 *	class CMaterialLexer
 *		- lexical analyzer of shader / material file language (css like)
 */
class CMaterialLexer {
public:
	/*
	 *	CMaterialLexer::<unnamed>
	 *		- token type values
	 */
	enum {
		token_White,
		token_Colon,
		token_Semicolon,
		token_LeftPar,
		token_RightPar,
		token_Begin,
		token_End, // it's important so operator tokens are in one piece
		token_Node,
		token_String,
		token_CData,
		token_Bool,
		token_Int,
		token_GLenum,
		token_Float,
		token_Pass,
		token_Texture1D,
		token_Texture2D,
		token_Texture3D,
		token_TextureCube, // it's important so texture tokens are in one piece
		token_URI,
		token_Reflect,
		token_Refract,
		token_Custom,
		token_DepthMap, // it's important so all other keyword tokens are in one piece
		token_EOF
	};

	/*
	 *	struct CMaterialLexer::TToken
	 *		- simple token structure for material file lexical analysis
	 */
	struct TToken {
	protected:
		int n_line, n_col;
		int n_type;
		union {
			int n_value;
			float f_value;
			char *p_s_value;
		} t_value;

	public:
		/*
		 *	CMaterialLexer::TToken::TToken()
		 *		- default constructor; sets line and column to 0 and type to token_EOF
		 */
		TToken();

		/*
		 *	CMaterialLexer::TToken::TToken(int _n_line, int _n_col, int _n_type)
		 *		- constructor for tokens with no data
		 *		- sets line to _n_line, column to _n_col and type to _n_type
		 */
		TToken(int _n_line, int _n_col, int _n_type);

		/*
		 *	CMaterialLexer::TToken::TToken(int _n_line, int _n_col, int _n_type, int n_value)
		 *		- constructor for tokens with integer data
		 *		- sets line to _n_line, column to _n_col and type to _n_type
		 *		- sets int value to n_value
		 */
		TToken(int _n_line, int _n_col, int _n_type, int n_value);

		/*
		 *	CMaterialLexer::TToken::TToken(int _n_line, int _n_col, int _n_type, float f_value)
		 *		- constructor for tokens with floating point data
		 *		- sets line to _n_line, column to _n_col and type to _n_type
		 *		- sets float value to f_value
		 */
		TToken(int _n_line, int _n_col, int _n_type, float f_value);

		/*
		 *	CMaterialLexer::TToken::TToken(int _n_line, int _n_col,
		 *		int _n_type, const char *p_s_value)
		 *		- constructor for tokens with string data
		 *		- sets line to _n_line, column to _n_col and type to _n_type
		 *		- sets string value to p_s_value (a new string is allocated and copied,
		 *		  check wheter allocation was succesful)
		 */
		TToken(int _n_line, int _n_col, int _n_type, const char *p_s_value);

		/*
		 *	CMaterialLexer::TToken::TToken(const TToken &r_t_token)
		 *		- copy-constructor
		 *		- note in case token is supposed to contain string data, a new string
		 *		  is allocated and copied, check wheter allocation was succesful
		 */
		TToken(const TToken &r_t_token);

		/*
		 *	CMaterialLexer::TToken::~TToken()
		 *		- default destructor; deallocates string data in case it was present
		 */
		~TToken();

		/*
		 *	bool CMaterialLexer::TToken::operator =(const TToken &r_t_token)
		 *		- copy-operator
		 *		- returns true on success, false in case there was
		 *		  not enough memory for string data
		 */
		bool operator =(const TToken &r_t_token);

		/*
		 *	inline int CMaterialLexer::TToken::n_Line() const
		 *		- returns zero-based index of line where token lies in source file
		 */
		inline int n_Line() const
		{
			return n_line;
		}

		/*
		 *	inline int CMaterialLexer::TToken::n_Column() const
		 *		- returns zero-based index of column where token starts in source file
		 */
		inline int n_Column() const
		{
			return n_col;
		}

		/*
		 *	inline int CMaterialLexer::TToken::n_Type() const
		 *		- returns token type
		 */
		inline int n_Type() const
		{
			return n_type;
		}

		/*
		 *	inline int CMaterialLexer::TToken::n_Value() const
		 *		- returns token integer value
		 *		- note this should be osed on tokens of type token_Int or token_Bool only
		 */
		inline int n_Value() const
		{
			_ASSERTE(n_type >= token_Bool && n_type <= token_GLenum);
			return t_value.n_value;
		}

		/*
		 *	inline float CMaterialLexer::TToken::f_Value() const
		 *		- returns token float value
		 *		- note this should be osed on tokens of type token_Float only
		 */
		inline float f_Value() const
		{
			_ASSERTE(n_type == token_Float);
			return t_value.f_value;
		}

		/*
		 *	inline const char *CMaterialLexer::TToken::p_s_Value() const
		 *		- returns token string value
		 *		- note this should be used on tokens of type token_Node,
		 *		  token_GLenum, token_String or token_CData only
		 */
		inline const char *p_s_Value() const
		{
			_ASSERTE(n_type >= token_Node && n_type <= token_CData);
			return t_value.p_s_value;
		}

#ifdef MATERIAL_LEXER_TOKEN_DUMP
		/*
		 *	void CMaterialLexer::TToken::Dump() const
		 *		- prints token contents
		 */
		void Dump() const;
#endif //MATERIAL_LEXER_TOKEN_DUMP
	};

protected:
	TToken m_t_token;
	bool m_b_token_valid;

	class CTokenEmit;

	struct TDrivingTable {
		struct TTransition {
			unsigned char n_char_min, n_char_max;
			char n_state;
		};

		struct TState {
			const TTransition *p_transition;
			char n_regexp_id;
			char n_transition_num;
		};
	};

	static TDrivingTable::TState m_p_state[33];
	static const TDrivingTable::TTransition m_p_transition[78];

	CScanner<char, TDrivingTable::TState,
		TDrivingTable::TTransition, CTokenEmit*> m_scanner;
	CTokenEmit *m_p_emit[8];

	class CFileReader;

	CFileReader *m_p_file;

	class CEmitSkip;
	class CEmitNode;
	class CEmitOperator;
	class CEmitString;
	class CEmitInt;
	class CEmitFloat;
	class CEmitGLenum;
	class CEmitCData;

public:
	/*
	 *	CMaterialLexer::CMaterialLexer(const char *p_s_filename)
	 *		- default constructor; creates a new parser, bound to p_s_filename
	 *		- call b_Status() to see wheter file was opened successfully
	 *		  (no parsing occurs at this moment)
	 *		- note all parsing occurs in us-ascii encoding
	 */
	CMaterialLexer(const char *p_s_filename);

	/*
	 *	CMaterialLexer::~CMaterialLexer()
	 *		- default destructor
	 */
	~CMaterialLexer();

	/*
	 *	bool CMaterialLexer::b_Status() const
	 *		- returns status of parser
	 *		- returns true if no errors occured, returns false if
	 *		  constructor or parsing failed
	 */
	bool b_Status() const;

	/*
	 *	bool CMaterialLexer::Get_NextToken()
	 *		- parses next token from file
	 *		- returns true on success, false on parsing / memory failure
	 *		- note reading past the end of the file is not an error
	 *		- always fails in case b_Status() returns false
	 */
	bool Get_NextToken();

	/*
	 *	inline bool CMaterialLexer::Get_NextToken(int n_accept_type)
	 *		- parses next token from file
	 *		- returns true on success and only if token type is equal
	 *		  to n_accept_type, otherwise returns false
	 *		- note reading past the end of the file is not an error
	 *		  if n_accept_type is token_EOF
	 *		- always fails in case b_Status() returns false
	 */
	inline bool Get_NextToken(int n_accept_type)
	{
		if(!Get_NextToken())
			return false;
		return m_t_token.n_Type() == n_accept_type;
	}

	/*
	 *	int CMaterialLexer::n_Cur_Line() const
	 *		- returns zero-based index of line where the parser last stopped
	 *		- this is useful for printing error messages on where parsing failed
	 */
	int n_Cur_Line() const;

	/*
	 *	int CMaterialLexer::n_Cur_Column() const
	 *		- returns zero-based index of column where the parser last stopped
	 *		- this is useful for printing error messages on where parsing failed
	 */
	int n_Cur_Column() const;

	/*
	 *	const CMaterialLexer::TToken &CMaterialLexer::t_Cur_Token() const
	 *		- returns constant reference to current token
	 *		- note in case no token was read, this returns token_EOF at line 0, col 0
	 *		- return value is undefined in case b_Status() returns false
	 */
	const TToken &t_Cur_Token() const;

	/*
	 *	inline int CMaterialLexer::n_Cur_Token() const
	 *		- returns type of current token
	 *		- note in case no token was read, this returns token_EOF
	 *		- return value is undefined in case b_Status() returns false
	 */
	inline int n_Cur_Token() const
	{
		return t_Cur_Token().n_Type();
	}
};

/*
 *	class CGLenumTable
 *		- small class for converting string values
 *		  of OpenGL enums to their numeric form
 */
class CGLenumTable {
protected:
	static const struct TTable {
		const char *p_s_name;
		unsigned short n_value;
	} m_p_table[];

public:
	/*
	 *	static int CGLenumTable::n_Value(const char *p_s_name)
	 *		- returns value of OpenGL enum p_s_name
	 *		- returns -1 in case p_s_name is not found in the table
	 */
	static int n_Value(const char *p_s_name);
};

#endif //__MATERIAL_SYSTEM_INCLUDED
